--[[
	This mod provides several choices on how to alter the way how the tech research cost ramps up in the tech tree.
	Scroll below to RP_COST_FUNCTION_TO_USE in order to choose how the cost should be altered.

	By default, a sigmoid function (limited growth) is applied.
	This means early tech ramp up slowly in RP cost (e.g. 50 RP becomes 60 RP, 150 RP becomes only 70 RP),
	but mid game tech ramp up faster (e.g. 2000 RP becomes 1450 RP, 2750 RP remains the same, 3150 becomes 3700 RP),
	and late game tech ramping up slower again (e.g. 7500 RP becomes 17000 RP but 10000 RP only becomes 21000 RP).

    The exp_increase function adds a moderate exponential ramp up in RP cost.
	This means early techs are not affected that much (e.g. 50 RP becomes 60 RP, an increase of 20%),
	but late game tech will become much more expensive (e.g. 25000 RP becomes 41000 RP, an increase of 64%).

	The exp_decrease function is similar to exp_increase but decreases RP cost instead.
	This means early tech are not affected as much (e.g. 50 RP becomes 40 RP, a decrease of 20%),
	but late game tech will become much more cheap (e.g. 25000 RP becomes 15000 RP, a decrease of 60%)

	Approach:
	Due to limited arithmetic support there is no native power-function that we could use here. 
	Custom implementations are possible but tricky. 
	Using a natural number exponent such as 2 for squaring would obviously be too extreme. 
	So this mod takes the 'lazy' approach and explicitly lists the new research cost values instead.
	These hardcoded values were computed in a separate script which is also shipped with the mod:
	generate_cost_map_output.py is a script written in Python 3.8. It uses the matplotlib library to plot the result,
	but you can just comment out or delete those lines and still run it. The script outputs the cost-map declaration
	of a chosen (or custom) function so that it can be directly copied into this LUA script below.
	
	Computation:
	See generate_cost_map_output.py for details.
	Note that (for mostly aesthetic reasons), all results are rounded to the nearest steps of 10,
	results higher than 1000 RP are rounded to nearest steps of 50,
	results higher than 10000 RP are rounded to nearest steps of 1000.
	An input research cost of 0 will always be output as 0 as it does not appear in the tech tree and could potentially
	conflict with other scripts or mods if it were changed.

	Compatibility:
	If you want to use custom values, either compute the new value and put an entry in the cost_map below or use the
	generate_cost_map_output.py script and enter your own custom function there.
	Computing dynamic values e.g. using TECH_VAR is best done after applying this mod by using modorder. 
	Note that new cost values of higher than 65536 RP will cause issues e.g. in ship design screen.
	
	Last update of this mod: 24th August 2021; Tested with version 1.50.13 improved
	
	Questions etc. to: romanhable@web.de (Epirasque)
]]


-- Add double-minus to a line to comment it out; remove the double-minus in the line with the cost function you want to use.
--RP_COST_FUNCTION_TO_USE='sigmoid'
--RP_COST_FUNCTION_TO_USE='exp_increase'
--RP_COST_FUNCTION_TO_USE='exp_decrease'
RP_COST_FUNCTION_TO_USE='custom'



cost_map_sigmoid = {
	[0] = 0,
	[50] = 50,
	[80] = 60,
	[150] = 70,
	[200] = 90,
	[250] = 100,
	[300] = 110,
	[400] = 150,
	[600] = 220,
	[650] = 240,
	[800] = 320,
	[900] = 370,
	[1150] = 540,
	[1500] = 840,
	[2000] = 1450,
	[2750] = 2750,
	[3150] = 3700,
	[3500] = 4600,
	[3750] = 5300,
	[4500] = 7650,
	[5000] = 9300,
	[6000] = 13000,
	[6500] = 14000,
	[7500] = 17000,
	[10000] = 21000,
	[12500] = 23000,
	[13500] = 23000,
	[15000] = 24000,
	[20000] = 25000,
	[25000] = 25000,
	[35000] = 25000
}

cost_map_exp_increase = {
    [0] = 0,
	[50] = 60,
	[80] = 100,
	[150] = 190,
	[200] = 260,
	[250] = 330,
	[300] = 400,
	[400] = 540,
	[600] = 830,
	[650] = 900,
	[800] = 1100,
	[900] = 1250,
	[1150] = 1650,
	[1500] = 2150,
	[2000] = 2950,
	[2750] = 4100,
	[3150] = 4700,
	[3500] = 5250,
	[3750] = 5650,
	[4500] = 6850,
	[5000] = 7650,
	[6000] = 9250,
	[6500] = 10000,
	[7500] = 12000,
	[10000] = 16000,
	[12500] = 20000,
	[13500] = 22000,
	[15000] = 24000,
	[20000] = 33000,
	[25000] = 41000,
	[35000] = 59000
}

cost_map_exp_decrease = {
    [0] = 0,
	[50] = 40,
	[80] = 60,
	[150] = 120,
	[200] = 150,
	[250] = 190,
	[300] = 230,
	[400] = 300,
	[600] = 440,
	[650] = 470,
	[800] = 570,
	[900] = 640,
	[1150] = 810,
	[1500] = 1050,
	[2000] = 1350,
	[2750] = 1850,
	[3150] = 2100,
	[3500] = 2350,
	[3750] = 2500,
	[4500] = 2950,
	[5000] = 3250,
	[6000] = 3900,
	[6500] = 4200,
	[7500] = 4800,
	[10000] = 6300,
	[12500] = 7800,
	[13500] = 8400,
	[15000] = 9250,
	[20000] = 12000,
	[25000] = 15000,
	[35000] = 21000
}

cost_map_custom = {
	[0] = 0,
	[50] = 50,
	[80] = 80,
	[150] = 150,
	[200] = 200,
	[250] = 250,
	[300] = 300,
	[400] = 400,
	[600] = 600,
	[650] = 650,
	[800] = 800,
	[900] = 900,
	[1150] = 1150,
	[1500] = 1500,
	[2000] = 2000,
	[2750] = 2750,
	[3150] = 3150,
	[3500] = 3500,
	[3750] = 3750,
	[4500] = 4500,
	[5000] = 5000,
	[6000] = 6000,
	[6500] = 6500,
	[7500] = 7500,
	[10000] = 10000,
	[12500] = 12000,
	[13500] = 14000,
	[15000] = 15000,
	[20000] = 20000,
	[25000] = 25000,
	[35000] = 35000
}


cost_map = nil;

if RP_COST_FUNCTION_TO_USE == 'exp_increase' then
	cost_map = cost_map_exp_increase
	msgbox('TECH_FKT: Changing RP cost of tech fields using \nexponential increase')
elseif RP_COST_FUNCTION_TO_USE == 'exp_decrease' then
    cost_map = cost_map_exp_decrease
	msgbox('TECH_FKT: Changing RP cost of tech fields using \nexponential decrease')
elseif RP_COST_FUNCTION_TO_USE == 'sigmoid' then
    cost_map = cost_map_sigmoid
    msgbox('TECH_FKT: Changing RP cost of tech fields using \nsigmoid function (limited growth)')
elseif RP_COST_FUNCTION_TO_USE == 'custom' then
    cost_map = cost_map_custom
    msgbox('TECH_FKT: Changing RP cost of tech fields using \ncustom function')
else
	msgbox('TECH_FKT: ERROR: Invalid cost function name!')
end

--default was 10000; here we always assume a value of 35000 after the increment, so pow(35000, 1.05) - pow(25000, 1.05) -> ~18000
hyper_advanced_tech_field_cost_increment_value = 18000 

advf = {
    advf_biology = 1,
    advf_power = 1,
    advf_physics = 1,
    advf_construction = 1,
    advf_fields = 1,
    advf_chemistry = 1,
    advf_computers = 1,
    advf_sociology = 1
}

function set_id(t, tag)
    for i,v in pairs(t) do
        v[tag] = i
    end
end

function make_branches(ff, keep_gray_boxes)
    local branches = {}
    local f2br = {}
    local starting_fields = rget_starting_fields()
	local nr_capped = 0

    local map = { engineering = "advanced_engineering" }
    if keep_gray_boxes then
        map.chemistry = "advanced_metallurgy"
        map.nuclear_fission = "advanced_fusion"
        map.physics = "fusion_physics"
        map.electronics = "optronics"
    end
    for i,s in pairs(starting_fields) do
        if map[s] ~= nil then
            starting_fields[i] = map[s]
        end
    end

    for i,f in pairs(starting_fields) do
        local n = rawlen(branches) + 1
        local b = {
            branch_last_field = "advf_" .. i,
            branchnum = n,
            field = f,
        }
        table.insert(branches, b)
        ff[f].is_branch_root = 1
    end

    for i,b in ipairs(branches) do
        local fname = b.field
        local marks = {}
        local n = 1
        while advf[fname] == nil and ff[fname] ~= nil do
            local f = ff[fname]
            if marks[fname] ~= nil then
                error(tostring("detected tech loop ", marks))
            end
            marks[fname] = n
            n = n + 1
            f.branch = i
            --f.techs = {}
			fname = f.next_field
        end
    end

    return branches
end

function changeFieldCost(fieldToChange)
	local oldCost = fieldToChange.cost
	if oldCost == nil then
		return 0
	end
    if cost_map[oldCost] ~= nil then
	    fieldToChange.cost = cost_map[oldCost]
	else
	    msgbox('TECH_FKT: WARNING: Previous cost value of ' .. tostring(oldCost) .. ' not covered by mod, will not be changed')
    end
end

function go()
    set_seed()
    local ff = rget_fields() set_id(ff, "name")
    local tt = rget_techs()  set_id(tt, "name")
	
	local nr_capped = 0
	for _, field in pairs(ff) do
		changeFieldCost(field)
	end
	
    rset_fields(ff)
    rset_techs(tt)
		
end

go()